import typer
import base64
from typing import List

cli = typer.Typer()


RESERVED_WORDS: List[str] = ["playground", "test", "testing", "privileged"]

@cli.command()
def encode_json_header(
        json: str = typer.Option(..., prompt=True, help="JSON string to encrypt"),
):
    json_bytes = json.encode('utf-8')
    base64_bytes = base64.b64encode(json_bytes)
    base64_str = base64_bytes.decode('utf-8')
    typer.echo(base64_str)


def prune_reserved_words(openapi_data: dict):
    for endpoint_path, endpoints in openapi_data['paths'].items():
        remove: List[str] = []
        for endpoint_verb, endpoint in endpoints.items():
            for reserved_word in RESERVED_WORDS:
                if 'tags' in endpoint:
                    if reserved_word in endpoint['tags']:
                        remove.append(endpoint_verb)
                else:
                    print("WARNING: An endpoint is missing a tag.")
        for remove_verb in remove:
            del endpoints[remove_verb]



import math
from typing import List
import typer
import pytest
from pytest import ExitCode
import os
from pathlib import Path
import coverage
import pylint.lint

cli = typer.Typer()

LINTING_CONFIG_VERSION: str = "0.0.1"

LINTING_THRESHOLD: float = 8.0
UNIT_TEST_COVERAGE_THRESHOLD: int = 69
INTEGRATION_TEST_COVERAGE_THRESHOLD: int = 50

OMIT_PATTERNS_LINTING: List[str] = [
    "example/**/*",
    "db/alembic/**/*",
    "service_meta.py",
    "service_template/**/*"
]

OMIT_PATTERNS_TESTING: List[str] = [
    "src/example/*",
    "src/db/alembic/*",
    "src/db/*schema*.py",
    "src/db/*model*.py",
    "src/*seeders*",
    "src/service_meta.py",
    "src/service_template/*",
    "src/db/__init__.py",
    "src/service.py"
]


def lint() -> float:
    omit_files: List[str] = []

    for omit_pattern in OMIT_PATTERNS_LINTING:
        for path in Path('src').glob(omit_pattern):
            omit_files.append(str(path))

    files_to_lint: List[str] = []

    for path in Path('src').rglob("*.py"):
        if str(path) not in omit_files:
            files_to_lint.append(str(path))

    results = pylint.lint.Run(
        [
            "-r",  # generate report
            "y",  # yes
            "-s",  # generate code score
            "y",  # yes
            f"--rcfile=.pylintrc.ehelply.{LINTING_CONFIG_VERSION}",
        ] + files_to_lint,
        exit=False
    )

    linting_score = results.linter.stats.global_note

    return linting_score


def test_base(tests_dir: Path, test_results_dir: Path) -> float:
    test_results_dir.mkdir(parents=True, exist_ok=True)

    result: ExitCode = pytest.main(
        [
            "-s",
            "-v",
            "--cov-report", f"html:{str(test_results_dir)}",
            "--cov=src",
            f"{str(tests_dir)}/"
        ]
    )

    if result != ExitCode.OK and result != ExitCode.NO_TESTS_COLLECTED:
        raise typer.Exit(code=result)

    cov = coverage.Coverage()
    cov.load()

    coverage_amount: float = cov.report(
        show_missing=True,
        omit=OMIT_PATTERNS_TESTING
    )

    return coverage_amount


def unit_test() -> float:
    root_path: Path = Path(os.getcwd())

    tests_dir: Path = Path(root_path).resolve().joinpath('tests').joinpath('unit')

    test_results_dir: Path = Path(root_path).resolve().joinpath('test-results').joinpath('unit')

    return test_base(tests_dir, test_results_dir)


def integration_test():
    root_path: Path = Path(os.getcwd())

    tests_dir: Path = Path(root_path).resolve().joinpath('tests').joinpath('integration')

    test_results_dir: Path = Path(root_path).resolve().joinpath('test-results').joinpath('integration')

    return test_base(tests_dir, test_results_dir)


def migrate():
    # Run DB migrations or unit tests for features relying on DB migrations will fail
    #   Thus, failing prod build/deploys
    typer.echo("Running Migrations..")
    from alembic.config import Config
    import alembic.command

    config = Config('alembic.ini')
    config.attributes['configure_logger'] = False

    alembic.command.upgrade(config, 'head')


@cli.command()
def linting(
        ignore_final_report: bool = typer.Option(False),
        ignore_failures: bool = typer.Option(False)
):
    report: List[str] = [
        "\n",
        "REPORT"
    ]

    """
    LINTING
    """
    linting_score: float = lint()

    if linting_score < LINTING_THRESHOLD and not ignore_failures:
        raise Exception(
            f"Linting score is {round(linting_score, 2)} which is below {LINTING_THRESHOLD}. Thus, build has failed."
        )

    report.append(f" * Linting score: {round(linting_score, 2)}/10, Linting threshold: {LINTING_THRESHOLD}")

    """
    REPORTING
    """

    report.append("\n")

    if not ignore_final_report:
        for line in report:
            typer.echo(line)


@cli.command()
def units(
        ignore_final_report: bool = typer.Option(False),
        ignore_failures: bool = typer.Option(False)
):
    report: List[str] = [
        "\n",
        "REPORT"
    ]

    migrate()
    report.append(" * Database Migrations were run")

    """
    UNIT TESTING
    """

    coverage_amount: float = unit_test()
    if math.ceil(coverage_amount) < UNIT_TEST_COVERAGE_THRESHOLD and not ignore_failures:
        raise Exception(
            f"Unit test coverage is {math.ceil(coverage_amount)}% which is below {UNIT_TEST_COVERAGE_THRESHOLD}%. Thus, build has failed."
        )

    report.append(
        f" * Unit test coverage: {int(coverage_amount)}%/100%, Unit test coverage threshold: {UNIT_TEST_COVERAGE_THRESHOLD}%")

    """
    REPORTING
    """

    report.append("\n")

    if not ignore_final_report:

        for line in report:
            typer.echo(line)


@cli.command()
def integrations(
        ignore_final_report: bool = typer.Option(False),
        ignore_failures: bool = typer.Option(False)
):
    report: List[str] = [
        "\n",
        "REPORT"
    ]

    migrate()
    report.append(" * Database Migrations were run")

    """
    INTEGRATION TESTING
    """

    coverage_amount: float = integration_test()
    if math.ceil(coverage_amount) < INTEGRATION_TEST_COVERAGE_THRESHOLD and not ignore_failures:
        raise Exception(
            f"Integration test coverage is {math.ceil(coverage_amount)}% which is below {INTEGRATION_TEST_COVERAGE_THRESHOLD}%. Thus, build has failed."
        )

    report.append(
        f" * Integration test coverage: {int(coverage_amount)}%/100%, Integration test coverage threshold: {INTEGRATION_TEST_COVERAGE_THRESHOLD}%")

    """
    REPORTING
    """

    report.append("\n")

    if not ignore_final_report:
        for line in report:
            typer.echo(line)



from ehelply_bootstrapper.integrations.integration import Integration
from ehelply_bootstrapper.utils.state import State
from ehelply_bootstrapper.utils.cryptography import Encryption

from ehelply_microservice_library.utils.paginate import Page
from ehelply_microservice_library.integrations.fact import get_fact_stage

from pydantic import BaseModel

import requests

from typing import Optional, Dict
import time
from jose import jwk, jwt
from jose.utils import base64url_decode

from ehelply_microservice_library.utils.exceptions import *


class M2mKey(BaseModel):
    secret_token: str
    participant_uuid: Optional[str] = None


class M2M(Integration):
    """
    M2M handles microservice to microservice communication
    """

    def __init__(self, service_gatekeeper_key: str) -> None:
        super().__init__("m2m")
        # Special requests object setup to talk to other services with auth
        self.requests: requests.Session = requests.Session()

        # Cognito public keys
        self.keys: list = []

        # M2M keys for speedy M2M auth
        self.m2m_keys: Dict[str, M2mKey] = {}

        self.enc: Encryption = Encryption([service_gatekeeper_key.encode(Encryption.STRING_ENCODING)])

        self.app_client_id: str = get_fact_stage("cognito-ehelply")['app_client_id']
        self.client_app_client_ids: str = get_fact_stage("cognito-ehelply")['client_app_client_ids']

    def init(self):
        # Get public keys for our user pool from Cognito
        region: str = get_fact_stage("cognito-ehelply")['region']
        userpool_id: str = get_fact_stage("cognito-ehelply")['user_pool_id']
        key_url: str = 'https://cognito-idp.{region}.amazonaws.com/{userpool_id}/.well-known/jwks.json'.format(
            region=region, userpool_id=userpool_id)

        response = requests.get(key_url)
        self.keys = response.json()['keys']

        # self.verify_token("eyJraWQiOiJwK2lwM1wvWlc5aXcyZGdjXC80Wmpkem16WDZcL2FsZUdyb1pLQVVaSmt4dEdzPSIsImFsZyI6IlJTMjU2In0.eyJjdXN0b206dXVpZCI6IjhjNjE4YzE0LTRmNmQtMTFlYS1hMGVhLTAyNmFkOTM4MzRlMCIsImF0X2hhc2giOiJweHZ2QlMzR1Q0VzNiSWI5czJNUGRBIiwic3ViIjoiZDUyNjBhNDMtNTFmZi00MGM4LWI5ZjgtNTBlZTBhNGFlMTg2IiwiY29nbml0bzpncm91cHMiOlsiY2EtY2VudHJhbC0xXzAzY0U4Tzcxb19Hb29nbGUiXSwiZW1haWxfdmVyaWZpZWQiOnRydWUsImN1c3RvbTpmaXJzdG5hbWUiOiJTaGF3biIsImlzcyI6Imh0dHBzOlwvXC9jb2duaXRvLWlkcC5jYS1jZW50cmFsLTEuYW1hem9uYXdzLmNvbVwvY2EtY2VudHJhbC0xXzAzY0U4TzcxbyIsImNvZ25pdG86dXNlcm5hbWUiOiJHb29nbGVfMTA3NDU5MTU2MzkyNjEzOTc5MjExIiwibm9uY2UiOiJjX0FHUkJfanZKSFd1cm55TjBqUm5DaUJoVXBQQ0JYQzlaRFQ2RExZR25ia2t2R2tYakIwNlVWZk51NjNLRVU5c3ZWNGtfdy1CV2o0dDNKQnRUclVsTGJhZ3lVc0F3QVpsMHRFVC01RnlmMWVjRmFCQ0xtYUVxY0E5YjNVc2xpUEJEOHdOTHBrZ3pjZVJ1STFGaWNKTGFHMTVEZmtUY0hfZ1dEMERfSVNXWEUiLCJjdXN0b206cGljdHVyZSI6Imh0dHBzOlwvXC9saDMuZ29vZ2xldXNlcmNvbnRlbnQuY29tXC9hLVwvQU9oMTRHajFaWXdQeEs3dGxnWUlLQkozZkVYY2pNN1pzbDlwdldNcGlDbnBvZz1zOTYtYyIsImF1ZCI6Ijc4dG80OGE5ZHZpMTNsamJvNGp0dmxnMG1kIiwiaWRlbnRpdGllcyI6W3sidXNlcklkIjoiMTA3NDU5MTU2MzkyNjEzOTc5MjExIiwicHJvdmlkZXJOYW1lIjoiR29vZ2xlIiwicHJvdmlkZXJUeXBlIjoiR29vZ2xlIiwiaXNzdWVyIjpudWxsLCJwcmltYXJ5IjoidHJ1ZSIsImRhdGVDcmVhdGVkIjoiMTU4MTcxMTEwMTYwMiJ9XSwidG9rZW5fdXNlIjoiaWQiLCJhdXRoX3RpbWUiOjE1ODM0NDEyNzUsIm5hbWUiOiJTaGF3biIsImV4cCI6MTU4MzQ0NDg3NSwiaWF0IjoxNTgzNDQxMjc1LCJlbWFpbCI6InNoYXduLmxhdmF3YXRlckBnbWFpbC5jb20ifQ.L90OonJLZjUMMlqOP5QxTFTTXKuGgkYGh-6AbZpZxN92y3G-B0G8IDEHpt5emkBvdH8ew5Gbu6a3ZvuFH1RKoHjqx19RZXQDmXWcr5k_4ONhQwoA3PesjVqUu_TRhfDHJM7n3Zy9h4C35W36PUJ0olaXDEtj2jDixAmA6ZbkdGWqFD2nWvzXAsts5ARoGs-3rLsF5heMkyOpJNRIrzZFQJOqUzw9CQRu3mpa6U3oIcl84lz9kFhVmcSKMJhtsNLW8C3-KFNeIBH4ikUY0LNKafQf7gtyVP4lh_ktiOfoWrbnhlX0NGN3zaghSI9p1XH4AmvOU6QghSGB3AAZR95Icw")

        try:
            secret_token: str = State.config.m2m.auth.secret_key
            access_token: str = State.config.m2m.auth.access_key

            if len(secret_token) == 0 or len(access_token) == 0:
                State.logger.warning("M2M credentials are not set")

            secret_token = self.enc.decrypt_str(secret_token.encode(Encryption.STRING_ENCODING))
            access_token = self.enc.decrypt_str(access_token.encode(Encryption.STRING_ENCODING))

            # Add to m2m_keys to allow for speedy M2M auth
            self.m2m_keys[access_token] = M2mKey(secret_token=secret_token)

            # Setup M2M communication with an access key and secret key for this service
            self.requests.headers.update({
                'X-Secret-Token': secret_token,
                'X-Access-Token': access_token
            })
        except:
            State.logger.severe("M2M credentials are invalid. Ensure they are encrypted.")

    def load(self):
        super().load()

    def get_speed_m2m_auth_participant(self, access_token: str) -> Optional[str]:
        if access_token in self.m2m_keys:
            return self.m2m_keys[access_token].participant_uuid
        return None

    def speedy_m2m_auth(self, access_token: str, secret_token: str) -> bool:
        """
        If a request is made with a known M2M key to this service, then there is no reason to do a bunch of API
            requests to security, access, and monitor to validate that request.

        Args:
            access_token:
            secret_token:

        Returns:

        """
        if access_token in self.m2m_keys and self.m2m_keys[access_token].secret_token == secret_token:
            return True
        return False

    def search(self, url: str, item_model=None, params: dict = None, headers: dict = None, page: int = 1,
               page_size: int = 25) -> Page:
        """
        Automatic pagination using M2M
        :param url:
        :param item_model:
        :param params:
        :param headers:
        :param page:
        :param page_size:
        :return:
        """
        page_params: dict = {
            'page': page,
            'page_size': page_size
        }

        if params is None:
            params = {}

        page_params.update(params)

        params = page_params

        page: Page = Page(**self.requests.get(url, params=params, headers=headers).json())

        if item_model:
            page.transform(lambda t: item_model(**t))

        return page

    def verify_token(self, token: str) -> dict:
        """
        Verify a Cognito token and return any claims
        :param token:
        :return:
        """

        # get the kid from the headers prior to verification
        headers = jwt.get_unverified_headers(token)
        kid = headers['kid']

        # search for the kid in the downloaded public keys
        key_index = -1

        for i in range(len(self.keys)):
            if kid == self.keys[i]['kid']:
                key_index = i
                break

        if key_index == -1:
            raise PublicKeyNotFound
            # print('Public key not found in jwks.json')
            # return False

        # construct the public key
        public_key = jwk.construct(self.keys[key_index])

        # get the last two sections of the token,
        # message and signature (encoded in base64)
        message, encoded_signature = str(token).rsplit('.', 1)

        # decode the signature
        decoded_signature = base64url_decode(encoded_signature.encode('utf-8'))

        # verify the signature
        if not public_key.verify(message.encode("utf8"), decoded_signature):
            raise SignatureVerificationFailure
            # print('Signature verification failed')
            # return False

        # print('Signature successfully verified')
        # since we passed the verification, we can now safely
        # use the unverified claims
        claims = jwt.get_unverified_claims(token)

        # additionally we can verify the token expiration
        if time.time() > claims['exp']:
            raise ExpiredToken
            # print('Token is expired')
            # return False

        # and the Audience  (use claims['client_id'] if verifying an access token)
        if claims['aud'] != self.app_client_id and claims['aud'] not in self.client_app_client_ids:
            raise TokenNotIssuedForAudience
            # print('Token was not issued for this audience')
            # return False

        # now we can use the claims
        # print(claims)
        return claims



"""
Constants having to do with Gatekeeper

.. data:: GATEKEEPER_KEY_ENVIRONMENT_NAME

    The name of the environment variable for the gatekeeper key

"""

GATEKEEPER_KEY_ENVIRONMENT_NAME: str = "EHELPLY_GATEKEEPER_KEY"
